#!/usr/bin/env python3

import sys
import socket
import struct

# default settings
host = "localhost"
port = 1337
if len(sys.argv) > 1:
    host, port = sys.argv[1], int(sys.argv[2])



#################################################################################
######################### START HELPER CODE #####################################
#################################################################################

p64 = lambda x: struct.pack("<Q", x)
u64 = lambda x: struct.unpack("<Q", x.rjust(8, b'\x00'))[0]

def dbg_print(txt):
    debug=False
    if debug:
        try:
            print(txt.decode('utf-8'))
        except UnicodeDecodeError:
            print(txt)

# globals
# using global variables is bad style, but we're h4xx0rs and it works
s = None
recv_buf = b""

# helper function to receive stuff
def recv_until(delim):
    global s
    global recv_buf
    assert s is not None

    if type(delim) is not bytes:
        delim = delim.encode('utf-8')
    tmp = recv_buf
    while delim not in tmp:
        try:
            r = s.recv(0x1000)
            if not r:
                dbg_print(tmp)
                return tmp
            tmp += r
        except InterruptedError:
            pass

    idx = tmp.find(delim)
    assert idx >= 0
    idx += len(delim)
    recv_buf = tmp[idx:]
    tmp = tmp[:idx]
    dbg_print(tmp)
    return tmp

def connect():
    global s
    s = socket.socket()
    s.connect((host, port))


def send(b):
    assert s is not None
    if type(b) is not bytes:
        b = b.encode('utf-8')
    s.send(b)

def register(name):
    global s
    assert s is not None
    recv_until(b" > ")
    s.send(b"r\n")

    recv_until(b" > ")
    s.send(name + b"\n")
    recv_until(b": ")
    uuid = recv_until(b"\n").strip()
    return uuid

def new_entry(uuid, data):
    global s
    assert s is not None

    datalen = len(data)

    recv_until(b" > ")
    s.send(b"l\n")

    recv_until(b"UUID:")
    s.send(uuid + b"\n")

    recv_until(b" > ")
    s.send(b"a\n")

    recv_until(b"Data length?")
    s.send(str(datalen).encode() + b"\n")

    recv_until(b"Data:\n")
    s.send(data)

    # logout
    recv_until(b" > ")
    s.send(b"l\n")

def delete_entry(uuid, idx):
    global s
    assert s is not None

    recv_until(b" > ")
    s.send(b"l\n")

    recv_until(b"UUID:")
    s.send(uuid + b"\n")
    recv_until(b" > ")
    s.send(b"d\n")

    l = recv_until(b"\n")
    if b"no entries" in l:
        print("[-] No entries")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return False

    s.send(str(idx).encode() + b"\n")

    d = recv_until(b" ")
    if b"Invalid" in d:
        print("[-] Invalid idx")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return False
    # logout
    recv_until(b" > ")
    s.send(b"l\n")
    return True

def modify_entry(lecturer_uuid, idx, data):
    global s
    assert s is not None

    datalen = len(data)

    recv_until(b" > ")
    s.send(b"l\n")

    recv_until(b"UUID:")
    s.send(lecturer_uuid + b"\n")

    recv_until(b" > ")
    s.send(b"m\n")

    l = recv_until(b"\n")
    if b"no entries" in l:
        print("[-] No entries")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return False

    s.send(str(idx).encode() + b"\n")

    d = recv_until(b" ")
    if b"Invalid" in d:
        print("[-] Invalid idx")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return False

    recv_until(b"Data length?")
    s.send(str(datalen).encode() + b"\n")

    recv_until(b"Data:\n")
    s.send(data)

    # logout
    recv_until(b" > ")
    s.send(b"l\n")
    return True


def read_entry(uuid, idx, decrypt):
    global s
    assert s is not None

    recv_until(b" > ")
    s.send(b"l\n")

    recv_until(b"UUID:")
    s.send(uuid + b"\n")

    recv_until(b" > ")
    s.send(b"r\n")

    l = recv_until(b"\n")
    if b"no entries" in l:
        print("[-] No entries")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return None
    s.send(str(idx).encode() + b"\n")
    d = recv_until(b" ")
    if b"Invalid" in d:
        print("[-] Invalid idx")
        # logout
        recv_until(b" > ")
        s.send(b"l\n")
        return None

    recv_until(b"[Y/N]? > ")
    end_token = b"===END ENTRY DATA===\n"
    if decrypt:
        s.send(b"Y\n")
        d = recv_until(b"\n");
        if b"===BEGIN ENTRY DATA===\n" not in d:
            print("[-] No begin", d)
            # logout
            recv_until(b" > ")
            s.send(b"l\n")
            return None
        r = recv_until(end_token)[:-len(end_token)]
    else:
        s.send(b"N\n")
        recv_until(b"===BEGIN ENCRYPTED ENTRY DATA===\n")
        r = recv_until(end_token)[:-len(end_token)]

    # logout
    recv_until(b" > ")
    s.send(b"l\n")
    return r
#################################################################################
######################### END HELPER CODE #####################################
#################################################################################

connect()

# example code show-casing the functions
message = b"this qualification task will be hard"
u1 = register(b"fkil")
print(f"[+] u1: {u1.decode()}")

print(f"[+] reading entry 1...")
print(f"[+] content of entry 1: {read_entry(u1, idx=1, decrypt=True)}")

new_entry(u1, message)
print(f"[+] created new entry...")
print(f"[+] reading entry 1...")
d = read_entry(u1, idx=1, decrypt=True)
assert d == message
print(f"[+] content of entry 1: {d}")
print(f"[+] modifying entry 1...")
modify_entry(u1, idx=1, data=message[:-4] + b"easy")
print(f"[+] reading entry 1...")
d = read_entry(u1, idx=1, decrypt=True)
print(f"[+] content of entry 1: {d}")
print(f"[+] deleting entry 1...")
delete_entry(u1, idx=1)

