## Goal of the task

The goal is to retrieve the flag by exploiting a bug in the `vuln` program. The flag is generally acquired by running the `/bin/get_flag` program on the remote system.

## Template code

To save you important time, we provide python wrapper functions to interact with the functionalities provided by the program. The helper code may (or may not) need to be modified by you.

Nevertheless, it can be useful to first manually interact with the program to get a better understanding of it.

You can connect to an hosted instance (remote or locally) by running `python client.py <host> <port>`

Note, in the course there will generally be no template code provided.

## Local setup

For binary exploitation it is generally very helpful to try out things locally first. Most importantly, this allows attaching a debugger, inspecting the state of the program and debug one's exploits.

We provide an **optional** Dockerfile. It should not be necessary to use it, but just in case you observe different behavior on the remote system compared to locally, you can mimic the environment of the hosted instance.

### Hosting challenge natively

We use `ynetd` (https://yx7.cc/code/) to host the challenge. You can host the challenge on a port on localhost via the following command (assuming `./vuln` and `./ynetd` are in the current working directory).

`./ynetd -p <port> ./vuln`

To aid in debugging and forward stderr and error messages to the network socket, you can additionally use `-se y`:

`./ynetd -se y -p <port> ./vuln`

Note, `vuln` assumes `/bin/get_flag` to be executable and will otherwise not retrieve a flag, so you may want to put a dummy get_flag binary (as provided) there.

### Docker

The Dockerfile bundles everything into an easily usable image, but debugging can be more complicated. Build it via:

`docker build -t bx25s:qual .`

and run the image via

`docker run --rm -dit --name bxqual -p <hostport>:1337 bx25s:qual`

the `--rm` flag will cause the container to be also deleted when stopped. To stop it, run:

`docker stop bxqual`

### Debugging

We recommend using `gdb`. You can attach to the program via `gdb -p $(pidof vuln)`. Depending on the system, this might require using elevated privilege. In the program there are two functions meant to be used as helpers in `gdb`, specifically `list_to_user` and `list_to_entry`. Make sure to look at the usage examples in the comments of `vuln.c`.

If you want to start debugging at a specific point in your exploit script, interrupt your exploit script using the python `input()` function and then attach with gdb. To continue your exploit script, give any input to your exploit script, i.e. press enter.
